using UnityEngine;
using DaggerfallWorkshop.Game;
using DaggerfallWorkshop.Game.Utility.ModSupport;
using DaggerfallWorkshop.Game.Utility.ModSupport.ModSettings; //required for mod settings
using Unity.Jobs;
using Unity.Burst;
using UnityEngine.Jobs;

public class BirdsInDaggerFall : MonoBehaviour {

    [System.Serializable]
    public class BirdData {
        public GameObject go;
        public int birdChance;
        public int minGroup;
        public int maxGroup;

        public BirdData(GameObject go, int birdChance, int minGroup, int maxGroup) {
            this.go = go;
            this.birdChance = birdChance;
            this.minGroup = minGroup;
            this.maxGroup = maxGroup;
        }
    }

    static BirdsInDaggerFall instance;

    public static BirdsInDaggerFall Instance {
        get { return instance ?? (instance = FindObjectOfType<BirdsInDaggerFall>()); }
    }

    public BirdData[] birdData;
    public int maxBirds = 50;
    public int minSpawnTime = 5;
    public int maxSpawnTime = 20;
    public int prewarmedBirdFlocks = 3;

    private GameObject mainCamera; //The mainCamera is used as the reference point to spawn birds
    private PlayerEnterExit playerEnterExit;
    private GameObject[] birdInstance;
    private bool wentInside;
    private int spawnID = 0;
    private float birdSpawnTimer = 1;
    private int totalBirdChance;
    private RaycastHit hitInfo;
    private bool prewarmed;

    public static Mod mod { get; private set; }

    [Invoke(StateManager.StateTypes.Start, 0)]
    public static void Init(InitParams initParams) {

        mod = initParams.Mod;  // Get mod
        new GameObject("Birds In DaggerFall").AddComponent<BirdsInDaggerFall>(); //Add Script to scene
        mod.IsReady = true;  // After finishing, set the mod's IsReady flag to true.
    }

    void Awake() {

        //If we loaded from mod, load setting
        if(mod != null) {
            ModSettings settings = mod.GetSettings();

            birdData = new BirdData[3];
            birdData[0] = new BirdData(mod.GetAsset<GameObject>("crow.Prefab", false),
            settings.GetInt("Crows", "CrowChance"),
            settings.GetInt("Crows", "MinCrows"),
            settings.GetInt("Crows", "MaxCrows"));

            birdData[1] = new BirdData(mod.GetAsset<GameObject>("Geese.Prefab", false),
            settings.GetInt("Geese", "GeeseChance"),
            settings.GetInt("Geese", "MinGeese"),
            settings.GetInt("Geese", "MaxGeese"));

            birdData[2] = new BirdData(mod.GetAsset<GameObject>("Hawk.Prefab", false),
            settings.GetInt("Hawks", "HawkChance"),
            settings.GetInt("Hawks", "MinHawks"),
            settings.GetInt("Hawks", "MaxHawks"));

            maxBirds = settings.GetInt("General", "MaxAmountofBirds");
            minSpawnTime = settings.GetInt("General", "MinSpawnTime");
            maxSpawnTime = settings.GetInt("General", "MaxSpawnTime");
            prewarmedBirdFlocks = settings.GetInt("General", "PrewarmedBirdFlocks");
        }

        for (int i = 0; i < birdData.Length; i++) {
            totalBirdChance += birdData[i].birdChance;
        }

        birdInstance = new GameObject[maxBirds];
        mainCamera = GameObject.FindGameObjectWithTag("MainCamera");
        playerEnterExit = GameManager.Instance.PlayerEnterExit;

        DaggerfallWorkshop.StreamingWorld.OnTeleportToCoordinates += OnPlayerTeleport;
    }

    private void OnPlayerTeleport(DaggerfallConnect.Utility.DFPosition worldPos) {

        //Clear our all existing birds when teleporting
        for (int i = 0; i < birdInstance.Length; i++) {
                if (birdInstance[i] != null)
                    Destroy(birdInstance[i]);
        }

        prewarmed = false;
    }


    void LateUpdate() {

        if (!GameManager.Instance.IsPlayingGame())
            return;

        //Check if the player is inside or has gone/left inside
        else if (playerEnterExit.IsPlayerInside && wentInside) {
            return;
        }
        else if (playerEnterExit.IsPlayerInside && !wentInside) {
            for (int i = 0; i < birdInstance.Length; i++) {
                if (birdInstance[i] != null)
                    birdInstance[i].SetActive(false);
            }
            wentInside = true;
            return;
        }
        else if (!playerEnterExit.IsPlayerInside && wentInside) {
            for (int i = 0; i < birdInstance.Length; i++) {
                if (birdInstance[i] != null)
                    birdInstance[i].SetActive(true);
            }
            wentInside = false;
        }

        //If the player arrives at a new location for the first time, spawn birds
        if(prewarmed == false) {

            //Check if the player is on the ground yet at the new location
            //This is the only reliable way to prevent birds from spawning early
            if (!Physics.Raycast(mainCamera.transform.position, -mainCamera.transform.up, 10.0f)) {
                return;
            }

            for (int i = 0; i < prewarmedBirdFlocks; i++) {

                int random = Random.Range(0, totalBirdChance);
                int result = 0;

                for (int j = 0; j < birdData.Length; j++) {
                    result += birdData[j].birdChance;

                    if (random < result) {
                        SpawnBirdFlock(j, 100);
                    }
                }
            }
            prewarmed = true;
            birdSpawnTimer = 0;
            return;
        }



        //Controll birds 
        for (int i = 0; i < birdInstance.Length; i++) {      
            if (birdInstance[i] != null) {
                birdInstance[i].transform.position += birdInstance[i].transform.forward * (Time.deltaTime * (8.0f + (i % 4.0f)));
            }
        }

        //spawn more birds
        birdSpawnTimer -= Time.deltaTime;

        if (birdSpawnTimer <= 0) {

            int random = Random.Range(0, totalBirdChance);
            birdSpawnTimer = Random.Range(minSpawnTime, maxSpawnTime);

            int result = 0;
            for(int i = 0; i < birdData.Length; i++) {
                result += birdData[i].birdChance;

                if(random < result) {
                    SpawnBirdFlock(i);                
                    return;
                }
            }
        }
    }

    private void SpawnBirdFlock(int birdID, float distanceFromPlayer = 150.0f) {

        transform.position = mainCamera.transform.position + (transform.up * Random.Range(25, 75));
        transform.rotation = Quaternion.Euler(0.0f, Random.Range(0.0f, 359.0f), 0.0f);

        //We do a raycast from the birdspawner location (above the player) to find it's starting point. If the raycast hit's something (like a mountain) then we don't spawn any birds there
        if (Physics.Raycast(transform.position, transform.forward, distanceFromPlayer*2)) {
            return;
        }

        //Move the spawnpoint to it's new location, a bit away from the player, then we give it a new direction in which the birds will fly, 
        //We want the birds to fly into the space where the player is, otherwise the player won't see the birds and it would be kind-of pointless
        transform.position += transform.forward * Random.Range(distanceFromPlayer, distanceFromPlayer*2);
        transform.eulerAngles += (120 + Random.Range(0, 80)) * Vector3.up;

        int numberOfBird = Random.Range(birdData[birdID].minGroup, birdData[birdID].maxGroup);

        for (int i = 0; i < numberOfBird; i++) {

            if (spawnID > birdInstance.Length - 1)
                spawnID = 0;

            //If we have gone through our list of birdspawners, we delete the old one and replace it with a new spawner. 
            if (birdInstance[spawnID] != null)
                Destroy(birdInstance[spawnID]);

            birdInstance[spawnID] = Instantiate(birdData[birdID].go,
                transform.position + new Vector3(Random.Range(-0.06f, 0.06f) * 100, Random.Range(-0.06f, 0.06f) * 100, Random.Range(-0.06f, 0.06f) * 100),
                transform.rotation) as GameObject;

            spawnID++;
        }
    }


    private void OnDestroy() {
        DaggerfallWorkshop.StreamingWorld.OnTeleportToCoordinates -= OnPlayerTeleport;
    }

    /*
     * 
    public struct MoveBird : IJobParallelForTransform {

        public void Execute(int index, TransformAccess transform) {
            transform.position += transform.for * (Time.deltaTime * (8.0f + (index % 4.0f)));
        }
    }
    */

}